
#include <cstdlib>
#include <cstdio>

void ensure_capacity(int);
int getAt(int);

/* simple implementation of an array based list to store integers */

int* arr;
int len;
int head;
int tail;

// check if list is empty (head is same as tail).
bool isempty(){
    return head == tail;
}

// returns number of elements in the list.
int size(){
    return (head - tail + len) % len;
}

// reallocate arr to ensure it can store at least n elements.
void ensure_capacity(int n){
    // e.g. double the list capacity, if needed.
    if(len > n)     // check if we need to do this...
        return;             // no, there's enough space there.
    int newlen = len;
    while(newlen <= n)    // until we can store n elements.
        newlen = newlen * 2;    // keep on doubling list length
    // allocate a new array of size newlen.
    int* newarr = new int[newlen];
    for(int i=0;i<size();i++)
        newarr[i] = getAt(i);       // copy old list into new list.
    delete[] arr;       // free old list.
    arr = newarr;       // setup new list.
    head = size();      // head is the end of the list.
    tail = 0;           // new list starts at 0.
    len = newlen;       // setup new length.
}


// get i-th element starting from tail (index 0).
int getAt(int i){
    if(i < 0)
        return arr[ (head + len + i) % len ];
    return arr[ (tail + i) % len ];
}

// set n at ith location.
void setAt(int i,int n){
    if(i < 0){
        arr[ (head + len + i) % len ] = n;
    }else{
        arr[ (tail + i) % len ] = n;
    }    
}

// push (add) an integer onto the list (at head).
void push(int n){
    ensure_capacity(size() + 1);    // ensure we have enough space.
    arr[head] = n;      // insert at head.
    head = (head + 1) % len;        // move head to next slot.
}

// reverse of push.
int pop(){
    if(size() <= 0)     // this is a *really* bad error. 
        return 0;       // we want to exit program if this happens.
    head = (head + len - 1) % len;  // slot before head index.
    return arr[head];           // return element under head.
}

// remove element from tail (similar to pop, except from other end).
int shift(){
    int n = arr[tail];
    tail = (tail + 1) % len;        // move head to next slot.
    return n;
}

// insert at the tail of the list (similar to push).
void unshift(int n){
    ensure_capacity(size() + 1);    // ensure we have enough space.
    tail = (tail + len - 1) % len;        // move tail 1 back.
    arr[tail] = n;      // insert at tail.
}

int main(){

    len = 8;
    arr = new int[len];
    head = tail = 0;

    for(int i=0;i<12;i++){

        if(i & 1){
            push(rand()%100);   // push 10 random numbers onto list.
            printf("push: %d): ",getAt(-1));
        }else{
            unshift(rand()%100);   // push 10 random numbers onto list.
            printf("unshift: %d): ",getAt(0));
        }

        printf("[");
        for(int j=0;j<size();j++)
            printf("%d,",getAt(j));
        printf("]\n");

    }

    // remove things from head of the list
    /*
    while(size() > 0){
        int n = pop();
        printf("pop: %d (len=%d, head=%d, tail=%d, size=%d)\n",n,len,head,tail,size());        
    } */
    
    /*
    // remove things from tail of the list
    while(size() > 0){
        int n = shift();
        printf("shift: %d (len=%d, head=%d, tail=%d, size=%d)\n",n,len,head,tail,size());        
    }
    */

    /*
    for(int i=0;i<size();i++){
        printf("element at: %d is: %d\n",i,getAt(i));
        printf("element at: %d is: %d\n",-i,getAt(-i));
    }
    */

    for(int i=0;i<size();i++){
        setAt(i,rand()%1000);       // change element in list.

        printf("[");        // display list.
        for(int j=0;j<size();j++)
            printf("%d,",getAt(j));
        printf("]\n");
    }

    return 0;
}
