/* 
 * econtacts.cpp
 * 
 * proram to manage contact emails;
 *   names and emails of folks.
 */


/*

 HW4: modify this code to list contacts in sorted order.
    in other words, modify the EContactList class and add a 
    "sort" function, that will sort the elements.

*/



#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <cctype>

// we need a structure (class) to store the actual
// data in...
class EContact {
    // private data of the class.
    char* name;
    char* email;
public:

    // default constructor
    EContact(){
        EContact("","");
    }

    // copy constructor.
    //   contact = contact 
    EContact(const EContact& c){
        // allocate memory for name.
        name = new char[ strlen(c.name)+1 ];
        strcpy(name,c.name);
        email = new char[ strlen(c.email)+1 ];
        strcpy(email,c.email);
    }

    // construct an object from name, and email.
    EContact(const char* n,const char* e){
        // allocate memory for name.
        name = new char[ strlen(n)+1 ];
        strcpy(name,n);
        email = new char[ strlen(e)+1 ];
        strcpy(email,e);
    }

    ~EContact(){
        delete[] name;
        delete[] email;
    }
    char* getName(){
        return name;
    }
    char* getEmail(){
        return email;
    }
    void setName(const char* s){
        delete[] name;
        name = new char[ strlen(s)+1 ];
        strcpy(name,s);
    }
    void setEmail(const char* s){
        delete[] email;
        email = new char[ strlen(s)+1 ];
        strcpy(email,s);
    }
    // assignment (e.g. = `operator');
    void set(const EContact& c){
        setName(c.name);
        setEmail(c.email);
    }
    EContact& operator=(const EContact& c){
        setName(c.name);
        setEmail(c.email);
        return *this;
    }
};

// remove trailing space (new lines, tabs, etc.) characters.
char* trim(char* s){
    int len = strlen(s);
    while(len > 0 && isspace(s[len-1])){
        s[len-1] = '\0';
        len--;
    }
    return s;
}

// prompt a user with str, if user types in anything,
// return that, else, return d (default).
char* prompt(const char* str,const char* d,char* mem,int maxlen){
    printf("%s",str);
    if(d != 0)
        printf(" [%s]",d);
    printf(": ");
    if(fgets(mem,maxlen,stdin) == 0)
        return (char*)d;
    int len = strlen(mem);  // get length of entered str.
    if(len <= 1)
        return (char*)d;
    /*
    if(mem[len-1] == '\n')  // if last character is newline
        mem[len-1] = '\0';  // clear it out.
    */
    return trim(mem);
}




EContact readContact(EContact* d){
    char name[100];
    char email[100];

    char* n = prompt("Name",d ? d->getName():NULL,name,sizeof(name));
    char* e = prompt("Email",d ? d->getEmail():NULL,email,sizeof(email));
    
    return EContact(n,e);
}

// display a contact with a number (index of contact).
void writeContact(EContact c,int n){
    printf("%d,%s,%s\n",n,c.getName(),
        c.getEmail());    
}






// int list (encapsulate the list code into a class).

class EContactList {

    EContact* arr;
    int len;
    int head;
    int tail;

    // reallocate arr to ensure it can store at least n elements.
    void ensure_capacity(int n){
        // e.g. double the list capacity, if needed.
        if(len > n)     // check if we need to do this...
            return;             // no, there's enough space there.
        int newlen = len;
        while(newlen <= n)    // until we can store n elements.
            newlen = newlen * 2;    // keep on doubling list length
        // allocate a new array of size newlen.
        EContact* newarr = new EContact[newlen];
        for(int i=0;i<size();i++)
            newarr[i] = get(i);       // copy old list into new list.
        delete[] arr;       // free old list.
        arr = newarr;       // setup new list.
        head = size();      // head is the end of the list.
        tail = 0;           // new list starts at 0.
        len = newlen;       // setup new length.
    }


public:

    EContactList(){
        len = 8;
        arr = new EContact[len];
        head = tail = 0;
    }

    // copy constructor (make a copy of an existing list)
    EContactList(EContactList& c){
        len = c.len;
        arr = new EContact[len];
        head = tail = 0;
        for(int i=0;i<c.size();i++)
            push(c.get(i));
    }

    ~EContactList(){
        // clean up whatever we allocated in constructor
        delete[] arr;
        len = head = tail = 0;  // zero things out.
    }

    // assignment operator.
    EContactList& operator=(EContactList& c){    
        delete[] arr;       // free list that's already there.
        len = c.len;
        arr = new EContact[len];
        head = tail = 0;
        for(int i=0;i<c.size();i++)
            push(c.get(i));
    }


    // check if list is empty (head is same as tail).
    bool isEmpty(){
        return head == tail;
    }

    // returns number of elements in the list.
    int size(){
        return (head - tail + len) % len;
    }

    // get i-th element starting from tail (index 0).
    EContact get(int i){
        if(i < 0)
            return arr[ (head + len + i) % len ];
        return arr[ (tail + i) % len ];
    }

    // set n at ith location.
    void set(int i,EContact n){
        if(i < 0){
            arr[ (head + len + i) % len ] = n;
        }else{
            arr[ (tail + i) % len ] = n;
        }    
    }

    /*
        // HW4: sort alphabetically.
    void sort(){

    }
    */


    // push (add) an integer onto the list (at head).
    void push(EContact n){
        ensure_capacity(size() + 1);    // ensure we have enough space.
        arr[head] = n;      // insert at head.
        head = (head + 1) % len;        // move head to next slot.
    }

    // reverse of push.
    EContact pop(){
        if(size() <= 0)     // this is a *really* bad error. 
            return arr[0];       // we want to exit program if this happens.
        head = (head + len - 1) % len;  // slot before head index.
        return arr[head];           // return element under head.
    }

    // remove element from tail (similar to pop, except from other end).
    EContact shift(){
        EContact n = arr[tail];
        tail = (tail + 1) % len;        // move head to next slot.
        return n;
    }

    // insert at the tail of the list (similar to push).
    void unshift(EContact n){
        ensure_capacity(size() + 1);    // ensure we have enough space.
        tail = (tail + len - 1) % len;        // move tail 1 back.
        arr[tail] = n;      // insert at tail.
    }

    // splice (index, length, replacewith).
    EContactList& splice(int n){
        // remove all elements after index n.
        while(size() > n)
            pop();
        return *this;
    }
    
    // at location n, remove l elements.
    EContactList& splice(int n,int l){
        EContactList tlst;
        // remove all elements after index n.
        while(size() > n)
            tlst.push(pop());   // save everything after n.
        while(l-- > 0)
            tlst.pop();     // remove l saved elements.
        // put back the remainer.
        while(tlst.size()>0)
            push(tlst.pop());
        return *this;
    }

    // at location n, remove l elements, replace with lst.
    EContactList& splice(int n,int l,EContactList& lst){
        //printf("before splice: len=%d, siz=%d, head=%d, tail=%d\n",len,size(),head,tail);
        EContactList tlst;
        // remove all elements after index n.
        while(size() > n)
            tlst.push(pop());   // save everything after n.
        while(l-- > 0)
            tlst.pop();     // remove l saved elements.
        for(int i=0;i<lst.size();i++)   // replacement list.
            push(lst.get(i));
        // put back the remainer.
        while(tlst.size()>0)
            push(tlst.pop());
            // printf("after splice: len=%d, siz=%d, head=%d, tail=%d\n",len,size(),head,tail);            
        return *this;
    }
    


};





// our program to read commands from user and 
// peform them. commands will include:
/// add
/// remove
/// list / sort
/// edit
/// find
int main(){
    int i;

    EContactList clst;
    

    char buf[1000];     // command buffer.
    printf("> ");
    // read next line (command) until end of file.
    while(fgets(buf,sizeof(buf),stdin)){
        trim(buf);
        //printf("you typed: [%s]\n",buf);

        // if user wants to get out... break this loop.
        if(!strcmp(buf,"exit") || !strcmp(buf,"quit"))
            break;

        // user wants to add a contact.
        if(!strcmp(buf,"add")){
            
            // read a contact...
            EContact c = readContact(NULL);
            
            // note, this contact is scoped 
            // to this if statement, once
            // it leaves this if statement,
            // the destructor gets called, 
            // and name/email are deleted.

            //printf("adding: %s, %s\n",c.getName(),c.getEmail());

            // add contact to "storage" array.
            clst.push(c);
        }

        if(!strcmp(buf,"list")){
            for(i=0;i<clst.size();i++)
                writeContact(clst.get(i),i);
        }


        if(!strcmp(buf,"remove")){
            char *p = prompt("Id of contact to remove",NULL,buf,sizeof(buf));
            if(p && isdigit(p[0])){
                int id = atoi(p);
                if(id >= 0 && id < clst.size()){
                    writeContact(clst.get(id),id);
                    prompt("Erase (y/n)","n",buf,sizeof(buf));
                    if(tolower(buf[0]) == 'y'){

                        // delete contact with id...
                        // put the last contact in the place we want to delete... and decrease the counter.
                        clst.splice(id,1);

                    }
                }
            }
        }



        if(!strcmp(buf,"edit")){
            char *p = prompt("Id of contact to edit",NULL,buf,sizeof(buf));
            if(p && isdigit(p[0])){
                int id = atoi(p);
                if(id >= 0 && id < clst.size()){
                    writeContact(clst.get(id),id);
                    EContact c = clst.get(id);
                    c = readContact(&c);
                    writeContact(c,id);
                    prompt("Save Changes? (y/n)","n",buf,sizeof(buf));
                    if(tolower(buf[0]) == 'y'){

                        // delete contact with id...
                        clst.set(id,c);
                    }
                }
            }
        }


        if(!strcmp(buf,"find")){
            char *p = prompt("search string",NULL,buf,sizeof(buf));
            if(p){
                // lowercase everything.
                for(i=0;i<strlen(p);i++)
                    p[i] = tolower(p[i]);
                
                for(i=0;i<clst.size();i++){
                    char buf2[1000];
                    strcpy(buf2,clst.get(i).getName());
                    strcat(buf2,clst.get(i).getEmail());
                    for(int j=0;buf2[j];j++)
                        buf2[j] = tolower(buf2[j]);

                    // printf("lookin for %s in %s\n",p,buf2);
                    if(strstr(buf2,p)) 
                    writeContact(clst.get(i),i);
                }
            }
        }

        printf("> ");
    }

    return 0;
}



