
#include <cstdlib>
#include <cstdio>

// int list (encapsulate the list code into a class).

class IntList {

    int* arr;
    int len;
    int head;
    int tail;

    // reallocate arr to ensure it can store at least n elements.
    void ensure_capacity(int n){
        // e.g. double the list capacity, if needed.
        if(len > n)     // check if we need to do this...
            return;             // no, there's enough space there.
        int newlen = len;
        while(newlen <= n)    // until we can store n elements.
            newlen = newlen * 2;    // keep on doubling list length
        // allocate a new array of size newlen.
        int* newarr = new int[newlen];
        for(int i=0;i<size();i++)
            newarr[i] = get(i);       // copy old list into new list.
        delete[] arr;       // free old list.
        arr = newarr;       // setup new list.
        head = size();      // head is the end of the list.
        tail = 0;           // new list starts at 0.
        len = newlen;       // setup new length.
    }


public:

    IntList(){
        len = 8;
        arr = new int[len];
        head = tail = 0;
    }

    // copy constructor (make a copy of an existing list)
    IntList(IntList& c){
        len = c.len;
        arr = new int[len];
        head = tail = 0;
        for(int i=0;i<c.size();i++)
            push(c.get(i));
    }

    ~IntList(){
        // clean up whatever we allocated in constructor
        delete[] arr;
        len = head = tail = 0;  // zero things out.
    }

    // assignment operator.
    IntList& operator=(IntList& c){    
        delete[] arr;       // free list that's already there.
        len = c.len;
        arr = new int[len];
        head = tail = 0;
        for(int i=0;i<c.size();i++)
            push(c.get(i));
    }


    // check if list is empty (head is same as tail).
    bool isEmpty(){
        return head == tail;
    }

    // returns number of elements in the list.
    int size(){
        return (head - tail + len) % len;
    }

    // get i-th element starting from tail (index 0).
    int get(int i){
        if(i < 0)
            return arr[ (head + len + i) % len ];
        return arr[ (tail + i) % len ];
    }

    // set n at ith location.
    void set(int i,int n){
        if(i < 0){
            arr[ (head + len + i) % len ] = n;
        }else{
            arr[ (tail + i) % len ] = n;
        }    
    }

    // push (add) an integer onto the list (at head).
    void push(EContact n){
        ensure_capacity(size() + 1);    // ensure we have enough space.
        arr[head] = n;      // insert at head.
        head = (head + 1) % len;        // move head to next slot.
    }

    // reverse of push.
    EContact pop(){
        if(size() <= 0)     // this is a *really* bad error. 
            return 0;       // we want to exit program if this happens.
        head = (head + len - 1) % len;  // slot before head index.
        return arr[head];           // return element under head.
    }

    // remove element from tail (similar to pop, except from other end).
    EContact shift(){
        int n = arr[tail];
        tail = (tail + 1) % len;        // move head to next slot.
        return n;
    }

    // insert at the tail of the list (similar to push).
    void unshift(int n){
        ensure_capacity(size() + 1);    // ensure we have enough space.
        tail = (tail + len - 1) % len;        // move tail 1 back.
        arr[tail] = n;      // insert at tail.
    }

    // splice (index, length, replacewith).
    IntList& splice(int n){
        // remove all elements after index n.
        while(size() > n)
            pop();
        return *this;
    }
    
    // at location n, remove l elements.
    IntList& splice(int n,int l){
        IntList tlst;
        // remove all elements after index n.
        while(size() > n)
            tlst.push(pop());   // save everything after n.
        while(l-- > 0)
            tlst.pop();     // remove l saved elements.
        // put back the remainer.
        while(tlst.size()>0)
            push(tlst.pop());
        return *this;
    }

    // at location n, remove l elements, replace with lst.
    IntList& splice(int n,int l,IntList& lst){
        //printf("before splice: len=%d, siz=%d, head=%d, tail=%d\n",len,size(),head,tail);
        IntList tlst;
        // remove all elements after index n.
        while(size() > n)
            tlst.push(pop());   // save everything after n.
        while(l-- > 0)
            tlst.pop();     // remove l saved elements.
        for(int i=0;i<lst.size();i++)   // replacement list.
            push(lst.get(i));
        // put back the remainer.
        while(tlst.size()>0)
            push(tlst.pop());
            // printf("after splice: len=%d, siz=%d, head=%d, tail=%d\n",len,size(),head,tail);            
        return *this;
    }
    


};



int main(){

    IntList lst;

    for(int i=0;i<12;i++){

        if(i & 1){
            lst.push(rand()%100);   // push 10 random numbers onto list.
            printf("push: %d): ",lst.get(-1));
        }else{
            lst.unshift(rand()%100);   // push 10 random numbers onto list.
            printf("unshift: %d): ",lst.get(0));
        }

        printf("[");
        for(int j=0;j<lst.size();j++)
            printf("%d,",lst.get(j));
        printf("]\n");
    }

    
    IntList lst2 = lst;      // call copy constructor.
    // if we're assigning upon creation, then we're calling
    // the copy constructor.

    IntList lst3;       // creates a *new* empty list.

    lst3 = lst2;        // does "assignment" (makes a memory copy of object); NOTE: By default, if object has pointers, then a copy of those pointers is also made... with result being that two objects are pointing to the same array.   To avoid that, we need to define our own assignment operator, to properly allocate a *new* array for the new list.




    printf("copy lst2: [");
    for(int j=0;j<lst2.size();j++)
        printf("%d,",lst2.get(j));
    printf("]\n");

    printf("copy lst3: [");
    for(int j=0;j<lst3.size();j++)
        printf("%d,",lst3.get(j));
    printf("]\n");

    printf("lst.splice(4,3): ");
    lst.splice(4,3);
    printf("lst: [");
    for(int j=0;j<lst.size();j++)
        printf("%d,",lst.get(j));
    printf("]\n");

    printf("lst.splice(4,3,lst2): ");
    lst.splice(4,3,lst2);
    printf("lst: [");
    for(int j=0;j<lst.size();j++)
        printf("%d,",lst.get(j));
    printf("]\n");
    
    return 0;
}
